use std::sync::Arc;
use std::sync::Mutex;
use crate::listener::WsListener;
use crate::upstream::UpStreamMgr;
use crate::common::types::DownStreamPack;
use crate::master::MasterMgr;
use std::sync::OnceLock;

pub struct UserApp {
    pub _masterMgr: MasterMgr,
    pub _upstream_mgr: UpStreamMgr,
    down_stream_msg_queue: Arc::<std::sync::Mutex::<r9::QueueLock<DownStreamPack>>>,
}

impl f9::app::UserApp for UserApp {

    fn get_pkg_name(&self) -> String {
        return "wsproxy".to_string();
    }

    fn init(&self) {
        MasterMgr::instance().init();
        WsListener::instance().init();
        UpStreamMgr::instance().init();
    }

    fn update(&self) {
        let mut work_is_empty = false;
        while !work_is_empty {
            let mut cur_node: Option<std::sync::Weak::<std::sync::Mutex::<DownStreamPack>>> = None;
            match self.fetch_work_list().lock() {
                Ok(work_list) => {
                    if !work_list.empty() {
                        let node = &work_list.first_entry();
                        cur_node = Some(node.clone());
                    } else {
                        work_is_empty = true;
                        break;
                    }
                }
                Err(poisoned) => {}
            }
            match cur_node {
                Some(v) => {
                    v.upgrade().unwrap().lock().unwrap().entry.lock().unwrap().del_init();
                    UpStreamMgr::instance().add_down_stream_pack(v.clone());
                }
                None => {}
            }
        }
        UpStreamMgr::instance().update();
    }

    fn uninit(&self) {
        WsListener::instance().uninit();
        UpStreamMgr::instance().uninit();
        MasterMgr::instance().uninit();
    }

    fn get_http_listen_port(&self) -> i32 {
        return 8080;
    }

}

static UserAppInstance: OnceLock<UserApp> = OnceLock::new();

impl UserApp {

    pub fn instance() -> &'static Self {
        return UserAppInstance.get_or_init(|| Self::new());
    }
    pub fn new() -> Self {
        let p = UserApp{
            _masterMgr: MasterMgr::new(),
            _upstream_mgr: UpStreamMgr::new(),
            down_stream_msg_queue: r9::QueueLock::<DownStreamPack>::new(),
        };
        return p;
    }

    fn fetch_work_list(&self) -> Arc::<Mutex::<r9::ListHeadLock<DownStreamPack>>> {
        self.down_stream_msg_queue.lock().unwrap().fetch();
        return self.down_stream_msg_queue.lock().unwrap().work_list.lock().unwrap().clone();
    }

    pub fn add_down_stream_pack(&self, socket_handle: u16, data: bytes::BytesMut) {
        println!("add_down_stream_pack1 len:{0}", data.len());
        let node = DownStreamPack::new(socket_handle, data);
        self.down_stream_msg_queue.lock().unwrap().push(&node.lock().unwrap().entry);
    }

}
