use std::collections::HashMap;
use actix::{Actor, StreamHandler};
use actix_web_actors::ws::{self, Message};
use actix_web::{
    web,
    App as ActixApp,
    HttpRequest,
    HttpResponse,
    HttpServer,
    Error
    };
use std::thread;
use actix::prelude::*;
use bytes::{BufMut, BytesMut};
use crate::app::UserApp;
use crate::constant;
use tokio::runtime::Runtime;
use crate::common::types::{DownStreamMessage};
use std::sync::OnceLock;
use std::sync::atomic::{AtomicU16, Ordering};

struct AppStateWithCounter {
    counter: std::sync::Mutex<i32>, // <- Mutex is necessary to mutate safely across threadsds
}

pub struct WsListener {
    socket_hash: std::sync::Mutex<HashMap<u16, Recipient<DownStreamMessage>>>,
    cur_socket_handle: AtomicU16,
    tokio_rt: Runtime,
}

static WsListenerInstance: OnceLock<WsListener> = OnceLock::new();

/// Define HTTP actor
pub struct WsConn {
    socket_handle: u16,
    recv_buf_len: usize,
    recv_buf: BytesMut,
}

impl Actor for WsConn {
    type Context = ws::WebsocketContext<Self>;
}

impl WsConn {
    fn parse_pkt(&mut self, bin: &actix_web::web::Bytes, ctx: &mut <WsConn as Actor>::Context) {
        let buf_len = bin.len();
        if buf_len > 0 {
            let mut already_read_bytes: usize = 0;
            {
                for i in 0..1024 * 64 * 2 {
                    self.recv_buf.put_i8(0);
                }
            }
            while true {
                let mut read_bytes = buf_len - already_read_bytes;
                if read_bytes > constant::MAX_PACKET_LEN - self.recv_buf_len {
                    read_bytes = constant::MAX_PACKET_LEN - self.recv_buf_len;
                }
                if read_bytes > 0 {
                    self.recv_buf[self.recv_buf_len..bin.len()].copy_from_slice(&bin);
                    self.recv_buf_len += read_bytes;
                    already_read_bytes += read_bytes;
                }

                let mut offset = 0;
                let mut pre_offect = 0;
                while true {
                    pre_offect = offset;
                    {
                        let pack_len = (self.recv_buf[0] as u16)+ ((self.recv_buf[1] as u16) << 8);
                        let msg_id = (self.recv_buf[2] as u16)+ ((self.recv_buf[3] as u16) << 8);
                        let seq_id = (self.recv_buf[4] as i32)+ ((self.recv_buf[5] as i32) << 8) +
                            ((self.recv_buf[7] as i32) << 16)+ ((self.recv_buf[7] as i32) << 24);
                        let magic_code = (self.recv_buf[8] as u16)+ ((self.recv_buf[9] as u16) << 8);
                        let ext_len = (self.recv_buf[10] as u16)+ ((self.recv_buf[11] as u16) << 8);
                        let mut tmp_bytes = BytesMut::with_capacity((pack_len + 24) as usize);
                        {
                            tmp_bytes.put_u16_le(pack_len);
                            tmp_bytes.put_u16_le(msg_id);
                            tmp_bytes.put_i32_le(seq_id);
                            tmp_bytes.put_u16_le(magic_code);
                            tmp_bytes.put_u16_le(self.socket_handle); //tmp_bytes.put_u16_le(ext_len);
                            tmp_bytes.put_i32_le(0);
                            tmp_bytes.put_i64_le(0);
                            for i in 24..tmp_bytes.capacity() {
                                tmp_bytes.put_i8(0);
                            }
                            tmp_bytes[24 .. (24+ pack_len as usize)].copy_from_slice(&bin[12..(12 + pack_len as usize)]);
                            println!("recv buf_len:{0}", tmp_bytes.len());
                        }
                        {
                            UserApp::instance().add_down_stream_pack(self.socket_handle, tmp_bytes);
                        }
                        offset += (pack_len  as usize) + 12;
                    }
                    if pre_offect >= offset || offset >= self.recv_buf_len {
                        break;
                    }
                }//end while
                if offset > 0 && offset < self.recv_buf_len {
                    self.recv_buf[..0].copy_from_slice(&bin);
                }
                self.recv_buf_len -= offset;
                if self.recv_buf_len > 0 {
                    self.recv_buf_len = 0;
                }
                if already_read_bytes >= bin.len() {
                    already_read_bytes = 0;
                    break;
                }
            }
        }
    }

}

impl StreamHandler<Result<Message, ws::ProtocolError>> for WsConn {

    fn started(&mut self, ctx: &mut Self::Context) {
        let  a: Recipient<DownStreamMessage> = ctx.address().recipient();
        WsListener::instance().on_connect(self, a);
    }
    fn handle(&mut self, item: Result<Message, ws::ProtocolError>, ctx: &mut Self::Context) {
        match item {
            Ok(Message::Text(text)) => {

            },
            Ok(Message::Ping(msg)) => {
                ctx.pong(&msg)
            },
            Ok(Message::Binary(bin)) => {
                self.parse_pkt(&bin, ctx);
            }
            Ok(Message::Close(reason)) => {
                ctx.close(reason);
                WsListener::instance().on_disconnect(self.socket_handle);
            },
            _ => {},
        }
    }
}

impl Handler<DownStreamMessage> for WsConn {
    type Result = ();

    fn handle(&mut self, msg: DownStreamMessage, ctx: &mut Self::Context) -> Self::Result {
        ctx.binary(msg.data);
    }
}

async fn index(req: HttpRequest, stream: web::Payload) -> Result<HttpResponse, Error> {
    let mut conn = WsConn{
        socket_handle: 0,
        recv_buf_len: 0,
        recv_buf: BytesMut::with_capacity(1024 * 64 * 2)
    };
    conn.recv_buf.resize(conn.recv_buf.capacity(), 0);
    let resp = ws::start(conn, &req, stream);
    println!("{:?}", resp);
    resp
}

impl WsListener {

    pub fn instance() -> &'static Self {
        return WsListenerInstance.get_or_init(|| Self::new());
    }

    pub fn new() -> Self {
        let p = Self{
            cur_socket_handle: AtomicU16::new(0),
            socket_hash: Default::default(),
            tokio_rt: tokio::runtime::Builder::new_multi_thread()
                .enable_all()
                .build()
                .unwrap(),
        };
        return p;
    }

    pub fn init(&self) {
        //self.work_thread = Arc::new(Some(1));
        {
           Some(thread::spawn(|| {
                println!("hello2");
                let mut rt = tokio::runtime::Builder::new_multi_thread()
                    .enable_all()
                    .build()
                    .unwrap();
                rt.block_on(async {
                    let counter = web::Data::new(AppStateWithCounter {
                        counter: std::sync::Mutex::new(0),
                    });
                    println!("hello2333");
                    HttpServer::new(move || {
                        // move counter into the closure
                        ActixApp::new() // <- register the created data
                            .app_data(counter.clone())
                            .route("/", web::get().to(index))
                    })
                        .bind(("0.0.0.0", 8080))?
                        .run()
                        .await
                });
            }));
            println!("hello3");
        }
    }

    pub fn uninit(&self) {

    }

    fn on_connect(&self, conn: &mut WsConn, socket: Recipient<DownStreamMessage>) {
        self.cur_socket_handle.fetch_add(1, Ordering::Relaxed);
        conn.socket_handle = self.cur_socket_handle.load(Ordering::Relaxed);
        self.socket_hash.lock().unwrap().insert(conn.socket_handle, socket);
        println!("on_connect socket_handle:{0}", conn.socket_handle);
    }

    fn on_disconnect(&self, socket_handle: u16) {
        self.socket_hash.lock().unwrap().remove(&socket_handle);
        println!("on_disconnect socket_handle:{0}", socket_handle);
    }

    pub fn send_to_downsteam(&self, socket_handle: u16, data: bytes::BytesMut) {
        match self.socket_hash.lock().unwrap().get(&socket_handle){
            Some(v) => {
                let msg = DownStreamMessage{
                    data: actix_web::web::Bytes::copy_from_slice(data.as_ref())
                };
                v.do_send(msg);
            },
            None => {

            }
        }
    }

}
